using System;
using System.Collections.Generic;
using System.Linq;
using System.Text.RegularExpressions;
using System.Threading.Tasks;
using AngleSharp.Html.Parser;
using LinqToDB;
using Microsoft.Extensions.Options;
using SunEngine.Core.Cache.CacheModels;
using SunEngine.Core.Cache.Services;
using SunEngine.Core.Configuration.Options;
using SunEngine.Core.DataBase;
using SunEngine.Core.Errors.Exceptions;
using SunEngine.Core.Models.Materials;
using SunEngine.Core.Security;
using SunEngine.Core.Services;
using SunEngine.Core.Utils;
using SunEngine.Core.Utils.TextProcess;


namespace SunEngine.Core.Managers
{
	public interface IMaterialsManager
	{
		ValueTask<int?> GetCategoryIdAsync(int materialId);
		ValueTask<int?> GetCategoryIdAsync(string materialName);
		Task<Material> GetAsync(int id);

		ValueTask CreateAsync(Material material, string tags, CategoryCached category,
			IReadOnlyDictionary<string, RoleCached> roles);

		ValueTask UpdateAsync(Material material, string tags, CategoryCached category,
			IReadOnlyDictionary<string, RoleCached> roles);

		/// <summary>
		/// Set IsDeleted = true
		/// </summary>
		Task DeleteAsync(Material material);

		/// <summary>
		/// Set IsDeleted = false
		/// </summary>
		Task RestoreAsync(Material material);

		Task DetectAndSetLastCommentAndCountAsync(int materialId);
		Task DetectAndSetLastCommentAndCountAsync(Material material);
		bool IsNameValid(string name);
		Task<bool> IsNameInDbAsync(string name);

		/// <summary>
		/// Move up in sort order (SortNumber field)
		/// </summary>
		ValueTask UpAsync(int id, int countMoves = 1);

		/// <summary>
		/// Move down in sort order (SortNumber field)
		/// </summary>
		ValueTask DownAsync(int id, int countMoves = 1);
	}

	public class MaterialsManager : DbService, IMaterialsManager
	{
		protected readonly ITagsManager tagsManager;
		protected readonly SanitizerService sanitizerService;
		protected readonly IOptionsMonitor<MaterialsOptions> materialsOptions;
		protected readonly ICategoriesCache categoriesCache;

		protected readonly Regex nameValidator =
			new Regex("^[a-zA-Z0-9-]{3," + DbColumnSizes.Materials_Name + "}$");

		public MaterialsManager(
			DataBaseConnection db,
			SanitizerService sanitizerService,
			ICategoriesCache categoriesCache,
			ITagsManager tagsManager,
			IOptionsMonitor<MaterialsOptions> materialsOptions) : base(db)
		{
			this.tagsManager = tagsManager;
			this.sanitizerService = sanitizerService;
			this.materialsOptions = materialsOptions;
			this.categoriesCache = categoriesCache;
		}


		public virtual async ValueTask<int?> GetCategoryIdAsync(int materialId)
		{
			return await db.Materials.Where(x => x.Id == materialId).Select(x => x.Category.Id)
				.FirstOrDefaultAsync();
		}

		public virtual async ValueTask<int?> GetCategoryIdAsync(string materialName)
		{
			return await db.Materials.Where(x => x.Name == materialName).Select(x => x.Category.Id)
				.FirstOrDefaultAsync();
		}

		public virtual Task<Material> GetAsync(int id)
		{
			return db.Materials.FirstOrDefaultAsync(x => x.Id == id);
		}

		public virtual async ValueTask CreateAsync(
			Material material,
			string tags,
			CategoryCached category,
			IReadOnlyDictionary<string, RoleCached> roles)
		{
			if (!roles.ContainsKey(RoleNames.Admin) || sanitizerService.SanitizerOptions.SanitizeAdminMaterials)
				material.Text = sanitizerService.Sanitize(new HtmlParser().ParseDocument(material.Text));

			material.SettingsJson = material.SettingsJson?.MakeJsonText();

			if (category.IsMaterialsSubTitleEditable)
				material.SubTitle = SimpleHtmlToText.ClearTags(sanitizerService.Sanitize(material.SubTitle));


			using (db.BeginTransaction())
			{
				material.Id = await db.InsertWithInt32IdentityAsync(material);
				await db.Materials.Where(x => x.Id == material.Id).Set(x => x.SortNumber, x => material.Id)
					.UpdateAsync();

				await tagsManager.MaterialCreateAndSetTagsAsync(material, tags);
				db.CommitTransaction();
			}
		}

		public virtual async ValueTask UpdateAsync(
			Material material,
			string tags,
			CategoryCached category,
			IReadOnlyDictionary<string, RoleCached> roles)
		{
			if (!roles.ContainsKey(RoleNames.Admin) || sanitizerService.SanitizerOptions.SanitizeAdminMaterials)
				material.Text = sanitizerService.Sanitize(new HtmlParser().ParseDocument(material.Text));

			material.SettingsJson = material.SettingsJson?.MakeJsonText();

			if (category.IsMaterialsSubTitleEditable)
				material.SubTitle = SimpleHtmlToText.ClearTags(sanitizerService.Sanitize(material.SubTitle));


			await db.UpdateAsync(material);

			await tagsManager.MaterialCreateAndSetTagsAsync(material, tags);
		}

		public virtual Task DeleteAsync(Material material)
		{
			return db.Materials.Where(x => x.Id == material.Id).Set(x => x.DeletedDate, DateTime.UtcNow).UpdateAsync();
		}

		public virtual Task RestoreAsync(Material material)
		{
			return db.Materials.Where(x => x.Id == material.Id).Set(x => x.DeletedDate, x => null)
				.UpdateAsync();
		}

		public virtual bool IsNameValid(string name)
		{
			return !int.TryParse(name, out _) && nameValidator.IsMatch(name);
		}

		public virtual Task<bool> IsNameInDbAsync(string name)
		{
			return db.Materials.AnyAsync(x => x.Name.ToLower() == name.ToLower());
		}

		public async ValueTask UpAsync(int id, int countMoves = 1)
		{
			var material = await db.Materials.Where(x => x.DeletedDate == null)
				.FirstOrDefaultAsync(x => x.Id == id);
			if (material == null)
				throw new SunEntityNotFoundException(nameof(Material), id);

			Material material2;

			if (countMoves == 1)
				material2 = await db.Materials
					.Where(x =>
						x.DeletedDate == null && x.CategoryId == material.CategoryId &&
						x.SortNumber > material.SortNumber)
					.OrderBy(x => x.SortNumber).FirstOrDefaultAsync();
			else
				material2 = await db.Materials.Where(x => x.DeletedDate == null &&
				                                          x.CategoryId == material.CategoryId &&
				                                          x.SortNumber > material.SortNumber + countMoves)
					.FirstOrDefaultAsync();
			if (material2 == null)
				throw new SunEntityNotFoundException(nameof(Material), "Previous material not found");

			db.BeginTransaction();

			long maxId = await db.Materials.MaxAsync(x => x.SortNumber) + 1;

			await db.Materials.Where(x => x.Id == material.Id)
				.Set(x => x.SortNumber, maxId).UpdateAsync();

			await db.Materials.Where(x => x.Id == material2.Id)
				.Set(x => x.SortNumber, material.SortNumber).UpdateAsync();

			await db.Materials.Where(x => x.Id == material.Id)
				.Set(x => x.SortNumber, material2.SortNumber).UpdateAsync();

			db.CommitTransaction();
		}

		public async ValueTask DownAsync(int id, int countMoves = 1)
		{
			var material = await db.Materials.Where(x => x.DeletedDate == null)
				.FirstOrDefaultAsync(x => x.Id == id);
			if (material == null)
				throw new SunEntityNotFoundException(nameof(Material), id);

			Material material2;

			if (countMoves == 1)
				material2 = await db.Materials
					.Where(x =>
						x.DeletedDate == null && x.CategoryId == material.CategoryId &&
						x.SortNumber < material.SortNumber)
					.OrderByDescending(x => x.SortNumber).FirstOrDefaultAsync();
			else
				material2 = await db.Materials.Where(x => x.DeletedDate == null
				                                          && x.CategoryId == material.CategoryId
				                                          && x.SortNumber == material.SortNumber + countMoves)
					.FirstOrDefaultAsync();

			if (material2 == null)
				throw new SunEntityNotFoundException(nameof(Material), "Next material not found");

			db.BeginTransaction();

			long maxId = await db.Materials.MaxAsync(x => x.SortNumber) + 1;

			await db.Materials.Where(x => x.Id == material.Id)
				.Set(x => x.SortNumber, maxId).UpdateAsync();

			await db.Materials.Where(x => x.Id == material2.Id)
				.Set(x => x.SortNumber, material.SortNumber).UpdateAsync();

			await db.Materials.Where(x => x.Id == material.Id)
				.Set(x => x.SortNumber, material2.SortNumber).UpdateAsync();

			db.CommitTransaction();
		}

		public virtual Task DetectAndSetLastCommentAndCountAsync(int materialId)
		{
			return DetectAndSetLastCommentAndCountAsync(db.Materials.FirstOrDefault(x => x.Id == materialId));
		}

		public virtual async Task DetectAndSetLastCommentAndCountAsync(Material material)
		{
			var commentsQuery = db.Comments.Where(x => x.MaterialId == material.Id && x.DeletedDate == null);

			var lastComment = await commentsQuery.OrderByDescending(x => x.PublishDate).FirstOrDefaultAsync();
			var lastCommentId = lastComment?.Id;
			var lastActivity = lastComment?.PublishDate ?? material.PublishDate;

			var commentsCount = await commentsQuery.CountAsync();

			await db.Materials.Where(x => x.Id == material.Id)
				.Set(x => x.CommentsCount, x => commentsCount)
				.Set(x => x.LastCommentId, x => lastCommentId)
				.Set(x => x.LastActivity, x => lastActivity)
				.UpdateAsync();
		}
	}
}